# forwardTime.py
#
# Bryan Daniels
# 10.13.2017
#
# Calculate the time at which worm speed changes from negative to positive.
#

from SloppyCell.ReactionNetworks import *
from SloppyCell.daskr import daeintException
import scipy.io
import scipy
import pylab
import numpy as np
from pylab import find
#import gatherResults
from toolbox.defaultFigure import prettyConfInt
from toolbox.binning import quantileBins

def modelForwardTime(m,laserIntensity,delay=10.,additionalIndepParams=[],
                     maxTime=100):
    """
    Given a SirIsaac model and laserIntensity, return the first
    time relative to the laser start at which the speed changes 
    from negative to positive (or None if this does not happen).
    """
    
    # set laser intensity, delay, and any other independent parameters
    indepParams = [laserIntensity,delay,] + additionalIndepParams
    m.setInitialVariables(indepParams)
    
    # set event that watches for positive speed
    # (remove any existing one to be safe)
    m.net.remove_component('positive speed')
    m.net.add_event('positive speed','gt(speed,0.)')
    
    # integrate trajectory
    try:
        traj = Dynamics.integrate(m.net,[0,maxTime])
    except daeintException:
        print "modelForwardTime WARNING: exception in model integration.  Returning None."
        return None

    # find the first time at which speed changes from negative to positive
    positiveSpeedEvents = filter(lambda e: e.event_id == 'positive speed',
                                 traj.events_occurred)
    if len(positiveSpeedEvents) > 0:
        # there was an event.  return the (laser-aligned) time
        return positiveSpeedEvents[0].time_fired - delay
    else:
        # there was no event
        return None

# 6.1.2018
def modelPeakSpeed(m,laserIntensity,delay=10.,timeRange=[0.,1.],numtimepoints=100,
    additionalIndepParams=[]):
    """
    Given a SirIsaac model and laserIntensity, return the peak
    (negative) speed.
    
    timeRange ([0.,1.])         : seconds post-laser
    """
    times = np.linspace(timeRange[0],timeRange[1],numtimepoints)
    try:
        speeds = m.evaluateVec(times+delay,'speed',
                               [laserIntensity,delay,] + additionalIndepParams)
    except daeintException:
        print "modelPeakSpeed WARNING: exception in model integration.  Returning None."
        return None
    return np.min(speeds)

def loadData(conditioni=0):
    """
    conditioni (0)      : 0 = control, 1 = ibuprofen, 2 = mutant
    """

    # () Load data (taken from initializeWormFitting.py)
    dataDir = '../Data/160325/'
    if conditioni == 0:
        matlabData = scipy.io.loadmat(dataDir+'control_oct.mat')
    elif conditioni == 1:
        matlabData = scipy.io.loadmat(dataDir+'ibuprofen.mat')
    elif conditioni == 2:
        matlabData = scipy.io.loadmat(dataDir+'mutant.mat')
    else:
        raise Exception

    laserIntensities = matlabData['I'][0] # (numTrials)
    smoothedSpeedsList = scipy.transpose(matlabData['fspeed'])
    times = scipy.arange(len(smoothedSpeedsList[0]))*1./60. # sec

    return laserIntensities,times,smoothedSpeedsList

# 10.16.2017
def pausedTrials(laserIntensities,times,smoothedSpeedsList,
                 thresholdMaxSpeed=10):
    """
    thresholdMaxSpeed (10)  : in pixels per second
    """
    # code taken from initializeWormFitting.py
    minTimeThreshold = 1. # sec
    maxTimeThreshold = 200./60. # sec
    thresholdTimeIndices = \
        filter(lambda i: minTimeThreshold <= times[i] <= maxTimeThreshold,
               range(len(times)))
    maxRevSpeed = scipy.amax(-smoothedSpeedsList[:,thresholdTimeIndices],axis=1)
    pauseStateList = maxRevSpeed < thresholdMaxSpeed
    pauseTrials = filter(lambda i: pauseStateList[i],range(len(laserIntensities)))
    return pauseTrials

def dataForwardTimes(conditioni=0,minTime=1.,removeNone=False,
                     minTimePositive=0.):
    """
    Return list of laser intensities and corresponding first times 
    relative to laser start at which the smoothed speed changes 
    from negative to positive (by default restricted to times after
    the laser start).
    
    conditioni (0)      : 0 = control, 1 = ibuprofen, 2 = mutant
    minTime (1.)        : (in seconds) Filter out any data before
                          minTime.  (Default filters out data before
                          laser start)
    removeNone (False)  : Optionally discard trials for which there
                          is no defined forward time
    minTimePositive (0.): (in seconds) Minimum duration of positive
                          values in order to count as moving forward.
                          (assumes equally spaced timesteps)
    """

    laserIntensities,times,smoothedSpeedsList = loadData(conditioni)

    # filter out pre-laser data
    timeIndices = find(times > minTime)
    times = times[timeIndices]
    smoothedSpeedsList = smoothedSpeedsList[:,timeIndices]

    # convert minTimePositive to number of timesteps
    # (assumes equally spaced timesteps)
    deltat = times[1]-times[0]
    minNumberAboveZero = int(minTimePositive/deltat)

    # calculate absolute linearly interpolated forward times
    forwardTimes = [ linearlyInterpolatedFirstPositiveDerivativeRoot(
                                times,speeds,minNumberAboveZero) \
                     for speeds in smoothedSpeedsList ]
                     
    # make times relative to laser start
    laserStart = 1 # sec
    relativeForwardTimes = []
    for t in forwardTimes:
        if t is not None:
            relativeForwardTimes.append(t - laserStart)
        else:
            relativeForwardTimes.append(None)

    # optionally discard trials for which there is no defined forward time
    if removeNone:
        indices = find([t != None for t in relativeForwardTimes])
        laserIntensities = laserIntensities[indices]
        relativeForwardTimes = scipy.array(relativeForwardTimes)[indices]

    return laserIntensities,scipy.array(relativeForwardTimes)

def linearlyInterpolatedFirstPositiveDerivativeRoot(times,values,minNumberAboveZero=1):
    """
    Simplistic algorithm for interpolating the first time at which
    the value changes from negative to positive.
    
    minNumberAboveZero (1)  : Minimum number of values above zero after the
                              crossing in order to count as a crossing
    """
    assert(len(times)==len(values))
    if len(find(scipy.array(values)==0))>0: raise Exception

    signs = scipy.sign(values)
    posDerivRootIndices = find((signs[1:]-signs[:-1]) == +2)
    if len(posDerivRootIndices) > 0:
        for i in posDerivRootIndices:
            if scipy.all(signs[i+1:i+1+minNumberAboveZero]>0) \
                and i<(len(signs)-minNumberAboveZero):
                return times[i] + (times[i+1]-times[i])/(1. - values[i+1]/values[i])
    
    return None

# see https://stackoverflow.com/questions/15556930/turn-scatter-data-into-binned-data-with-errors-bars-equal-to-standard-deviation
def plotBinnedData(x,y,nbins=10,dataLabel='',binnedLabel='',
                   scatterColor='C0',scatterFillColor='C0',
                   scatterSize=2,scatterEdgeWidth=0.3,
                   binnedColor='C1',binnedFillColor='C1',
                   removeNone=True,stdOfMean=True,
                   equalSpacing=True,**kwargs):
    """
    (The returned values are standard deviations; 
     Plotted error bars are standard deviation of the mean)
     
    equalSpacing (True)         : If True, use bins equally spaced with respect to
                                  x values.  
                                  If False, use quantile bins that have (close to)
                                  equal numbers of data points.
    """
    if removeNone:
        indices = find([yval != None for yval in y])
        x = scipy.array(x)[indices]
        y = scipy.array(y)[indices]
    
    if equalSpacing:
        bins = nbins
    else:
        bins = quantileBins(x,nbins,inclusiveEndpoints=False)
    
    n, bins = scipy.histogram(x, bins=bins)
    sy, bins = scipy.histogram(x, bins=bins, weights=y)
    sy2, bins = scipy.histogram(x, bins=bins, weights=y*y)
    mean = sy / n
    stddevOfMn = [ scipy.sqrt(sy2[i]/n[i] - mean[i]*mean[i]) / scipy.sqrt(n[i])   \
                for i in range(len(mean)) ]
    stddev = [ scipy.sqrt(sy2[i]/n[i] - mean[i]*mean[i])   \
                for i in range(len(mean)) ]
    if stdOfMean:
        std = stddevOfMn
    else:
        std = stddev

    print "plotBinnedData: n =",n

    pylab.plot(x,y,'o',mec=scatterColor,color=scatterFillColor,
               ms=scatterSize, label=dataLabel, mew=scatterEdgeWidth)
    pylab.errorbar((bins[1:]+bins[:-1])/2., mean, yerr=std,
                   mec=binnedColor,color=binnedFillColor,
                   ls='',capsize=2,marker='o',label=binnedLabel,
                   **kwargs)

    return stddev,bins

# 6.6.2018
def plotBinnedConfidenceInterval(x,y,nbins=10,label='',
                   ax=None,filterNone=True,**kwargs):
    """
    (90% confidence intervals)
    """
    if ax is None:
        ax = pylab.gca()
    
    n, bins = scipy.histogram(x, bins=nbins)
    
    # edge effects
    bins[0] -= 1e-9
    bins[-1] += 1e-9
    
    medians,lowers,uppers = [],[],[]
    for i,binedges in enumerate(zip(bins[:-1],bins[1:])):
        inBinIndices = filter(lambda idx: binedges[0] < x[idx] <= binedges[1],range(len(y)))
        xInBin = np.array(x)[inBinIndices]
        yInBin = np.array(y)[inBinIndices]
        if n[i] != len(yInBin):
            print xInBin,binedges,n[i],len(xInBin)
            raise Exception
        if filterNone:
            yInBin = filter(lambda yval: yval is not None,yInBin)
        medians.append(scipy.median(yInBin))
        lowers.append(scipy.percentile(yInBin,5))
        uppers.append(scipy.percentile(yInBin,95))
    
    #pylab.plot(x,y,'o',mec=scatterColor,color=scatterFillColor,
    #           ms=scatterSize, label=dataLabel, mew=scatterEdgeWidth)
    print n
    prettyConfInt(ax,(bins[1:]+bins[:-1])/2., medians, lowers, uppers,
                   **kwargs)
