# gatherResults.py
#
# Bryan Daniels
# 8.26.2016
#
# Automated system for combining fitting from multiple conditions, calculating
# out-of-sample goodness of fit, and producing plots and reports.
#
# Coming mostly from plot_results.ipynb
#

# Allow matplotlib plots without Xwindows
import matplotlib
matplotlib.use('Agg')

import sys, copy, time
from SirIsaac import *
from SirIsaac.fitAllParallel import *
from subprocess import call
import glob

from toolbox import defaultFigure,gifsicle
defaultFigure.setDefaultParams()
import pylab
from matplotlib import gridspec
from matplotlib.ticker import MaxNLocator

try:
    from pygraphviz import *
    NETWORKSUPPORT = True
except ImportError:
    NETWORKSUPPORT = False

SBMLSUPPORT = hasattr(IO,'to_SBML_file')

NAMES = ['Control','Ibuprofen','Mutant']
DEFAULTBINS = [0,70,110,180]
BINCOLORS = ['cornflowerblue','orange','C1','crimson','C4']
RECALCULATECOST = True #True #False
REMAKEREPORTS = True #True #True #False
STOPFITTINGN = 3 # now calculated using out-of-sample costs
#NUMDATA = [159,349,68] (number of trials with thresholdMaxSpeed = True)

def gifsicleSpeedVsTime(fileNumString,**kwargs):
    Nlist = scipy.sort(loadFitProbData(fileNumString).keys())
    fileList,annotateList = [],[]
    for N in Nlist:
         names = glob.glob(fileNumString+'_reports/'+fileNumString+\
                           '_numDatapoints'+str(N)+'*speed*')
         if len(names) == 1:
             fileList.append(names[0])
             annotateList.append("N = "+str(N))
    print "gifsicleSpeedVsTime: Found",len(fileList),"files."
    if len(fileList) > 0:
        return gifsicle.gifsicleAnnotate(fileList,annotateList,
                                         fileNumString+'_speedVsTime.gif',
                                         **kwargs)


def totalNumDatapoints(data):
    N = 0
    for dataSingleCondition in data:
        for species,d in dataSingleCondition.items():
            N += len(d.keys())
    return N

def updateGoodnessOfFit(fittingProblem,data=None,indepParamsList=None,
                        includePriors=False):
    """
    Measure goodness of fit, chiSq/DOF, for all models in given fittingProblem,
    and store in outOfSampleCostDict.
    
    Can also be a FittingProblemMultipleCondition instance,
    in which case data and indepParamsList should be lists, one for
    each condition.
    """
    # set up lists in the case of multiple conditions
    if hasattr(fittingProblem,'fittingProblemList'):
        fittingProblemList = fittingProblem.fittingProblemList
        if data is None: 
            data = [ None for fp in fittingProblemList ]
        if indepParamsList is None: 
            indepParamsList = [ None for fp in fittingProblemList ]
        dataList = data
        indepParamsListList = indepParamsList
    else:
        fittingProblemList = [fittingProblem]
        dataList = [data]
        indepParamsListList = [indepParamsList]

    # initialize outOfSampleCostDict if it's not there
    if not hasattr(fittingProblem,'outOfSampleCostDict'):
        fittingProblem.outOfSampleCostDict = {}

    # calculate cost only for models that have been fit in all conditions
    if len(fittingProblemList) == 1:
        fitModels = fittingProblemList[0].logLikelihoodDict.keys()
    else:
        fitModels = scipy.intersect1d([ fp.logLikelihoodDict.keys() \
                                        for fp in fittingProblemList ])
    # calculate cost only for models that have not yet been calculated
    modelNames = [ m for m in fitModels \
                   if m not in fittingProblem.outOfSampleCostDict.keys() ]

    # check that each model in each condition has been fit
    for fp in fittingProblemList:
        for modelName in modelNames:
            if modelName not in fp.logLikelihoodDict.keys():
                raise Exception, str(modelName)+" does not appear in logLikelihoodDict."\
                                                "  Has it been fit?"
    
#    # loop over conditions
#    for fp,d,ip in zip(fittingProblemList,dataList,indepParamsListList):
#        if d is None: d = fp.fittingData
#        if indepParamsList is None: indepParamsList = fp.indepParamsList
#
#        # loop over models, adding chiSqOverN for each condition
#        for modelName in modelNames:
#            m = fp.fittingModelDict[modelName]
#            try:
#                chiSq = 2. * m.currentCost(d,ip,includePriors=includePriors)
#            except Utility.SloppyCellException:
#                print "goodnessOfFit: WARNING: Integration failed for "+str(modelName)
#                chiSq = scipy.inf
#            chiSqDict[modelName] += chiSq
#            print modelName, chiSq

    # loop over models
    for modelName in modelNames:
        fittingProblem.outOfSampleCostDict[modelName] = \
            goodnessOfFit_model(fittingProblemList,dataList,
                                indepParamsListList,modelName,
                                includePriors=includePriors)
    

def numParameters(model):
    return len(model.getParameters())

def goodnessOfFit_model(fittingProblemList,dataList,indepParamsListList,
                        modelName,includePriors=False,storeChiSq=True,
                        useDOF=True):
    """
    useDOF (True)           : Divide by degrees of freedom 
                              = # datapoints - # parameters
                              instead of # datapoints
                              (6.5.2018)
    """
    # loop over conditions, adding chiSq for each
    chiSqSum = 0.
    numParametersSum = 0.
    for fp,d,ip in zip(fittingProblemList,dataList,indepParamsListList):
        m = fp.fittingModelDict[modelName]
        numParametersSum += numParameters(m)
        try:
            chiSq = 2. * m.currentCost(d,ip,includePriors=includePriors)
        except Utility.SloppyCellException:
            print "goodnessOfFit: WARNING: Integration failed for "+str(modelName)
            chiSq = scipy.inf
        chiSqSum += chiSq
        print modelName, chiSq
        
        if storeChiSq:
            # store chiSq for individual fittingProblem
            if not hasattr(fp,'outOfSampleCostDict'):
                fp.outOfSampleCostDict = {}
            if useDOF:
                nu = totalNumDatapoints(d) - numParameters(m)
            else:
                nu = totalNumDatapoints(d)
            fp.outOfSampleCostDict[modelName] = chiSq/nu

    # divide by the degrees of freedom /or/ total number of datapoints
    N = scipy.sum([ totalNumDatapoints(d) for d in dataList ])
    if useDOF:
        nu = N - numParametersSum
        print "N =",N,", numParametersSum =",numParametersSum
    else:
        nu = N
    return chiSqSum / nu

def makeDirIfNotThere(newDir,outputDir):
    if newDir not in os.listdir(outputDir):
        os.mkdir(outputDir+'/'+newDir)

def wormReport(fileNumString,N,modelName,outputDir='.'):
    #fp = fpd[N]
    fp = loadCombinedFittingProblem(fileNumString,N)

    if modelName is None:
        print "wormReport: modelName = None.  Skipping."
        return

    reportDirectory = fileNumString + "_reports"
    makeDirIfNotThere(reportDirectory,outputDir)

    filePrefix = "/" + fileNumString + "_numDatapoints"+str(N) + "_" + str(modelName)

    # speed vs time
    makeDirIfNotThere("speedVsTime",reportDirectory)
    speedVsTimePlot(fp,modelName,reportDirectory+"/speedVsTime"+filePrefix)

    # network figures
    if NETWORKSUPPORT:
        makeDirIfNotThere("networkFigures",reportDirectory)
        networkFigures(fp,modelName,reportDirectory+"/networkFigures"+filePrefix)

    # SBML output
    if SBMLSUPPORT:
        #SBMLoutput(fp,modelName,filePrefix)
        # assume we also have mathematica
        makeDirIfNotThere("streamPlots",reportDirectory)
        streamPlots(fp,modelName,reportDirectory+"/streamPlots"+filePrefix)

    # TeX output
    makeDirIfNotThere("tex",reportDirectory)
    texOutput(fp,modelName,reportDirectory+"/tex"+filePrefix)

    print "wormReport: Done with reports for numDatapoints =",N





def speedVsTimePlot(fp,modelName,filePrefix,plotHiddenNodes=True,names=NAMES,
                    saveToPDF=True,speedMin=-110,speedMax=25,x2min=-10,x2max=15,
                    relTimeMin=-1,relTimeMax=4,figsize=(12,6)):
    
    
    # choose inputs to plot
    delay = 10 #3 # seconds
    if fp.indepParamNames == ['laserIntensity','laserIntensity_delay']:
        laserIntensities = [0,50,100,150,200]
        #laserIntensities = [25,30,35,40,45]
        colors = ['darkblue','blue','cornflowerblue','orange','crimson']
        indepParamsList = [[i,delay] for i in laserIntensities]
    elif fp.indepParamNames == ['laserIntensity', 'pauseState', 'laserIntensity_delay']:
        laserIntensities = [0,50,100,150,200,
                            0,50,100,150,200]
        pauseStates = [0,0,0,0,0,
                       1,1,1,1,1]
        colors = ['darkblue','blue','cornflowerblue','orange','crimson',
                  'gray','gray','gray','gray','gray']
        indepParamsList = [[i,p,delay] for i,p in zip(laserIntensities,pauseStates)]
    else:
        raise Exception
    
    inputMin,inputMax = 0,max(laserIntensities)
    timeMin,timeMax = delay+relTimeMin,delay+relTimeMax

    pylab.figure(figsize=figsize)
    i = 0
    for f,name in zip(fp.fittingProblemList,names):
        i += 1
        if modelName is None: modelName = fp.maxLogLikelihoodName()
        m = f.getBestModel(modelName=modelName)
        if plotHiddenNodes:
            # indep params won't be plotted
            numSpecies = len([ n for n in m.speciesNames if n not in m.indepParamNames ])
            a = [ [pylab.subplot(numSpecies,3,3*j+i)] for j in range(numSpecies) ]
        else:
            numSpecies = 1 # assumes we've only got speed to plot
            ax2 = pylab.subplot(2,3,i)
            a = [[ax2]]
        for indepParams,color in zip(indepParamsList,colors):
            a = m.plotResults(f.fittingData,[indepParams],
                              maxTime=timeMax,color=color,
                              label=indepParams[0],
                              wspace=0.25,existingAxArray=a,
                              plotHiddenNodes=plotHiddenNodes)
        # dotted line at 0.5 indicating region of in-sample data
        ax1 = a[0][0]
        pylab.sca(ax1)
        #pylab.plot([0.5,0.5],[speedMin,speedMax],'k:')
        pylab.title(name)
        ax1.axis(ymin=speedMin,ymax=speedMax,xmin=timeMin,xmax=timeMax)
        if plotHiddenNodes:
            # set y bounds on hidden variable X_2
            a[1][0].axis(ymin=x2min,ymax=x2max,xmin=timeMin,xmax=timeMax)
            a[-1][0].axis(ymin=inputMin,ymax=inputMax,xmin=timeMin,xmax=timeMax)
        if i==1: 
            ax1.set_ylabel('Speed (px/sec)')
        else:
            [ ax[0].set_ylabel('') for ax in a ]
        if i==1 and plotHiddenNodes:
            pylab.sca(a[numSpecies-1][0])
            #leg = pylab.legend(loc=2,title='Laser int.',
            #                        labelspacing=0.15)
            pylab.ylabel('Laser intensity')
        else:
            leg = None
        for ax in a:
            pylab.sca(ax[0])
            #defaultFigure.makePretty(leg=leg)
        pylab.xlabel('Time (sec)')
        
    pylab.subplots_adjust(bottom=0.2,wspace=0.2,hspace=0.1)
    if saveToPDF:
        if plotHiddenNodes:
            pylab.savefig(filePrefix+'_speedVsTime_withHiddenNodes.pdf')
        else:
            pylab.savefig(filePrefix+'_speedVsTime.pdf')


def networkFigures(fp,modelName,filePrefix,names=NAMES,paramsList=None,
    display=True,laserScale=200,weightScale=-50,**kwargs):
    i = 0
    if display:
        pylab.figure(figsize=(12,3))
    filenameList = [filePrefix+'_network_figure_'+name for name in names]
    for f,filename,name in zip(fp.fittingProblemList,filenameList,names):
        
        if modelName is None: modelName = fp.maxLogLikelihoodName()
        m = f.getBestModel(modelName=modelName)
        
        if paramsList is not None:
            m.initializeParameters(paramsList[i])
        
        # 5.2.2016 temporarily remove the timeVarying species until pulseInput
        # is fixed to correctly update the models' netLists
        oldSpeciesNames = copy.deepcopy(m.speciesNames)
        oldInputNames = copy.deepcopy(m.indepParamNames)
        oldNumInputs = copy.deepcopy(m.numInputs)
        m.speciesNames = oldSpeciesNames[:-1]
        m.indepParamNames = oldInputNames[:-1]
        m.numInputs = oldNumInputs - 1
        
        if m.numInputs == 1: # no paused state as input
            indepParamMax = [laserScale]
        elif m.numInputs == 2: # paused state as input
            indepParamMax = [laserScale,1]
        f.networkFigureBestModel(filename,modelName=modelName,
                                 indepParamMax=indepParamMax,showWeights=False,
                                 weightScale=weightScale,**kwargs)
                                 # negative weight scale produces correct sign?
        
        # 5.2.2016 (see above)
        m.speciesNames = copy.deepcopy(oldSpeciesNames)
        m.indepParamNames = copy.deepcopy(oldInputNames)
        m.numInputs = copy.deepcopy(oldNumInputs)

        i += 1

        # display here in jupyter notebook
        if display:
            call(['convert',filename+'.eps',filename+'.png'])
            img=matplotlib.image.imread(filename+'.png')
            pylab.subplot(1,3,i)
            pylab.imshow(img)
            pylab.axis('off')
            pylab.title(name)
    if display:
        pylab.savefig(filePrefix+'_networkFigures.pdf')


def texOutput(fp,modelName,filePrefix,names=NAMES):
    # TO DO: XXXXXXXXXX
    # combine into a single PDF with a table for parameters in multiple conditions
    for conditioni,name in enumerate(names[:len(fp.fittingProblemList)]):
        IO.eqns_TeX_file(fp.fittingProblemList[conditioni].fittingModelDict[modelName].net,
                    filePrefix+'_'+str(name)+'.tex')

def streamPlots(fp,modelName,filePrefix,outputSBML=True):
    """
    Use Mathematica script to create pdfs containing "stream plots"
    and csvs listing locations of fixed points at zero laser input.
    """
    if outputSBML:
        SBMLoutput(fp,modelName,filePrefix)
    call(["./stream_plots_script.m",filePrefix])

def SBMLoutput(fp,modelName,filePrefix,names=NAMES,fixBug=True):
    """
    Note: Assumes condition names are the first ones that appear in 
    global variable NAMES.
    """
    for conditioni,name in enumerate(names[:len(fp.fittingProblemList)]):
        net = fp.fittingProblemList[conditioni].fittingModelDict[modelName].net
        
        if fixBug: # pulseInput bug affects older fpds
            s = net.species.get('laserIntensity_timeVarying')
            if s is not None:
                if type(s.compartment) != str:
                    s.compartment = 'Comp'
    
        IO.to_SBML_file(net,filePrefix+'_'+str(name)+'.sbml')

def readFixedPoints(fileNumString,numTimepoints,bestModelName):
    """
    Read in fixed point information from files written by streamPlots.
    """
    reportDirectory = fileNumString + "_reports"
    
    fixedPointsByName,fixedPointsStabilityByName = {},{}
    for i,name in enumerate(NAMES):
        filename = fileNumString + "_numDatapoints" + str(numTimepoints) + \
                   "_" + bestModelName + "_fixedPoints_" + name + ".csv"
        try:
            raw = scipy.loadtxt(reportDirectory+"/streamPlots/"+filename,
                                delimiter=',',dtype=str)
            if len(scipy.shape(raw)) == 1: raw = scipy.array([raw])
            if len(raw[0]) > 0:
                fixedPoints = scipy.array(raw[:,:-1],dtype=float)
                fixedPointsStability = scipy.array(raw[:,-1]=='True')
            else:
                fixedPoints = scipy.array([])
                fixedPointsStability = scipy.array([])
            fixedPointsByName[name] = fixedPoints
            fixedPointsStabilityByName[name] = fixedPointsStability
        except IOError:
            pass
    
    return fixedPointsByName,fixedPointsStabilityByName



def storeFixedPoints(fileNumString,numTimepoints,bestModelName,saveToDisk=True):
    """
    Read in fixed point information from files written by streamPlots
    and save within fittingProblem.
    
    Note: Assumes condition names are the first ones that appear in
    global variable NAMES.
    """
    #fp = fpd[numTimepoints]
    fp = loadCombinedFittingProblem(fileNumString,numTimepoints)
    
    # read fixed point csv
    fixedPoints,fixedPointsStability = readFixedPoints(fileNumString,
                                        numTimepoints,bestModelName)
    
    # store fixed point info in fp
    for i,conditionName in enumerate(NAMES):
      if fixedPoints.has_key(conditionName):
        fpCond = fp.fittingProblemList[i]
        if not hasattr(fpCond,'fixedPointsDict'):
            fpCond.fixedPointsDict = {}
            fpCond.fixedPointsStabilityDict = {}
        fpCond.fixedPointsDict[bestModelName] = \
            fixedPoints[conditionName]
        fpCond.fixedPointsStabilityDict[bestModelName] = \
            fixedPointsStability[conditionName]

    if saveToDisk:
        # save updated fp within directory structure
        Nfilename = directoryPrefixNonly(fileNumString,numTimepoints) \
            +'/'+loadFitProbData(fileNumString).values()[0]['saveFilename']
        save(fp,Nfilename)


def bestOutOfSampleModelName(fp,stopFittingN=None):
    lowestCost = scipy.inf
    besti,bestModelName = None,None
    for i,modelName in enumerate(fp.fittingModelNames): # (want correct order)
        cost = fp.outOfSampleCostDict.get(modelName,scipy.inf)
        if cost < lowestCost:
            besti = i
            bestModelName = modelName
            lowestCost = copy.copy(cost)

    if bestModelName is None:
        return None
    elif stopFittingN and besti+stopFittingN >= len(fp.outOfSampleCostDict.keys()):
        return None
    else:
        return bestModelName

def bestOutOfSampleNumStableFixedPoints(fp,stopFittingN=None):
    """
    List of number of stable fixed points over conditions for the 
    model with best out-of-sample predictions.
    """
    bestModelName = bestOutOfSampleModelName(fp,stopFittingN)
    if bestModelName is None: return [ None for f in fp.fittingProblemList ]
    
    numList = []
    for f in fp.fittingProblemList:
        if hasattr(f,'fixedPointsStabilityDict'):
            numList.append(scipy.sum(f.fixedPointsStabilityDict[bestModelName]))
        else: # fixed points not calculated or produced an error, or there are none
            numList.append(scipy.nan)

    return numList

# 2.9.2017 from gather_results.ipynb
def plotGoodnessOfFit(fpdCombined,chiSqMin=0.75,chiSqMax=2.0,Nmin=0,Nmax=None,
                      numInSampleTrials=None,stopFittingN=3,rescaleFactor=1.,
                      plotNumFixedPoints=True):
    """
    rescaleFactor (1.)          : Multiply all chiSq values by a constant.
    """
    Nlist = scipy.sort(fpdCombined.keys())
    bestChiSqList = []
    bestNumParamsList = []
    bestNumStableList = []
    NwithData = []
    for N in Nlist:
        fp = fpdCombined[N]
        if hasattr(fp,'outOfSampleCostDict'):
            bestModelName = bestOutOfSampleModelName(fp,stopFittingN)
            if bestModelName is not None:
                NwithData.append(N)
                bestChiSqList.append( fp.outOfSampleCostDict[bestModelName] * rescaleFactor )
                bestNumParamsList.append( fp.numParametersDict[bestModelName] )
                bestNumStableList.append( bestOutOfSampleNumStableFixedPoints(fp) )

    if len(NwithData) == 0:
        raise Exception, "No data subsets have best-fit models"

    pylab.figure(figsize=(5,3.125)) #(6,3.75)) #(6,5))
    if plotNumFixedPoints:
        gs = gridspec.GridSpec(2, 1, height_ratios=[3, 1])
    else:
        gs = gridspec.GridSpec(1, 1)
    pylab.subplots_adjust(hspace=0,left=0.15,right=0.775,bottom=0.125)
    
    # optionally divide by number of trials to get datapoints per trial
    if numInSampleTrials is not None:
        datapointsPerTrial = scipy.array(NwithData)/float(numInSampleTrials)
    else:
        datapointsPerTrial = NwithData
    if Nmax is None:
        Nmax = max(datapointsPerTrial)

    # print lengths of out of sample indices used
    if hasattr(fpdCombined[Nlist[0]],'outOfSampleCostIndices'):
        print "Number of trials in each out-of-sample dataset:"
        print [ len(indices) for indices in fpdCombined[Nlist[0]].outOfSampleCostIndices ]

    # () top subplot
    pylab.subplot(gs[0])
    
    # plot chiSq
    #pylab.scatter(NwithData,bestChiSqList,c=bestNumStableList)
    pylab.plot(datapointsPerTrial,bestChiSqList,
               ',-',color='orange',lw=2,
               label='$\\chi^2 /$ dof')
    if numInSampleTrials is not None:
        pylab.xlabel('Number of datapoints per trial')
    else:
        pylab.xlabel('Number of datapoints $N_D$')
    pylab.ylabel('$\\chi^2 /$ dof')
    pylab.axis(ymax=chiSqMax,ymin=chiSqMin,xmin=Nmin,xmax=Nmax)
    #pylab.yticks([1.5,2,2.5,3.])
    pylab.yticks([1.,1.5,2.])

    leg1 = pylab.legend(numpoints=1,loc=4,labelspacing=0.15,
                        borderaxespad=0.1)

    # plot number of parameters
    ax1 = pylab.gca()
    ax2 = ax1.twinx()
    pylab.plot(datapointsPerTrial,bestNumParamsList,'k-',lw=1,
               label='Num. parameters')
    pylab.ylabel('Number of\nmodel parameters')
    pylab.axis(ymin=0,xmin=Nmin,xmax=Nmax)
    ax2.yaxis.set_major_locator(MaxNLocator(integer=True)) # force integer ticks
    
    leg2 = pylab.legend(loc=1,labelspacing=0.15,
                        borderaxespad=0.1)

    if plotNumFixedPoints:
        pylab.gca().set_xticklabels([])
    defaultFigure.makePretty(leg1,ax=ax1)
    defaultFigure.makePretty(leg2,ax=ax2)

    if plotNumFixedPoints:
        # () bottom subplot
        pylab.subplot(gs[1])
        
        # plot number of stable fixed points at zero laser input
        bestNumStableList = scipy.array(bestNumStableList)
        for i in range(len(bestNumStableList[0])):
            name = NAMES[i]
            pylab.plot(datapointsPerTrial,bestNumStableList[:,i]+i*0.05,
                       '.-',lw=0.5,label=name)
        pylab.yticks([0,1,2])
        pylab.ylabel('# stable\n fixed pts.')
        pylab.axis(ymin=-0.3,ymax=2.3,xmin=Nmin,xmax=Nmax)
        if numInSampleTrials is not None:
            pylab.xlabel('Number of datapoints per trial')
        else:
            pylab.xlabel('Number of datapoints $N_D$')

        if len(bestNumStableList[0]) > 1:
            leg = pylab.legend(loc='center right',bbox_to_anchor=(1.35,0.5))
            defaultFigure.makePretty(leg)

    defaultFigure.makePretty()

    pylab.subplots_adjust(bottom=0.175,top=0.95)


def getOutOfSampleData(fileNumString,forceAllTrials=True):
    """
    As of 5.22.2017, hard-coded to always use all trials.  May change in the future.
    """
    
    saveFilename = loadFitProbData(fileNumString).values()[0]['saveFilename']
    allDataFilename = saveFilename[:-4]+'_allData.dat'
    allData = load(allDataFilename)

    # choose out-of-sample data indices (see notes 8.26.2016)
    if (len(allData.fittingProblemList) == 3) or forceAllTrials: # all 3 condition case
        testIndices = [ range(len(fp.fittingData)) for fp in allData.fittingProblemList ]
    elif len(allData.fittingProblemList) == 1: # control only case
        testIndices = [ range(100,159) ]
        raise Exception
    else:
        raise Exception

    dataList = [ [ fprob.fittingData[i] for i in indices ] for fprob,indices in zip(allData.fittingProblemList,testIndices) ]
    indepParamsListList = [ [ fprob.indepParamsList[i] for i in indices ] \
                           for fprob,indices in zip(allData.fittingProblemList,testIndices) ]

    return testIndices,dataList,indepParamsListList

# 2/13/2017
def binnedOutOfSampleVsTime(fileNumString,laserBins=DEFAULTBINS,
                            roundNum=4,forceAllTrials=False,stdOfMean=True,
                            returnAll=False):
    """
    Returns timesList,meanDataListList,stdDataListList.
    
    timesList has shape (#conditions)x(#times)
    mean and stdDataListList have shape (#conditions)x(#bins)x(#times)
    
    returnAll (False)           : Optionally return all individual trial data as
                                  a fourth return value, with same shape as mean
                                  and stdDataListList but each entry is a list
                                  (with varying length)
    """
    testIndices,dataList,indepParamsListList = getOutOfSampleData(fileNumString,
                                                                  forceAllTrials)
    
    timesList,meanDataListList,stdDataListList = [],[],[]
    if returnAll: allDataListList = []
    
    relativeTimesList = uniqueRelativeTimes(fileNumString,roundNum,forceAllTrials)
    
    # loop over conditions
    for data,indepParamsList,allRelativeTimes in \
        zip(dataList,indepParamsListList,relativeTimesList):
        
        laserBins = scipy.array(laserBins)
        meanDataList,stdDataList,allDataList = [],[],[]
        # loop over laser bins
        for binmin,binmax in zip(laserBins[:-1],laserBins[1:]):
            numPerBin = 0
            alignedData = [ [] for t in allRelativeTimes ]
            # loop over trials
            for d,ip in zip(data,indepParamsList):
              if binmin <= ip[0] < binmax:
                numPerBin += 1
                trialTimes = scipy.array(d['speed'].keys())
                trialRelativeTimes = trialTimes - ip[1]
                trialRoundedTimes = scipy.round_(trialRelativeTimes,roundNum)
                for i,t in enumerate(allRelativeTimes):
                    timeIndices = pylab.find(trialRoundedTimes == t)
                    if len(timeIndices)>0:
                        #trialTime = trialTimes[timeIndex[0]]
                        #alignedData[i].append(d['speed'][trialTime][0])
                        trialTime = trialTimes[timeIndices]
                        alignedData[i].extend([ d['speed'][time][0] for time in trialTime])
                    else: # time not found
                        pass
            print "binnedOutOfSampleVsTime: numPerBin =",numPerBin
            meanDataList.append( [ scipy.mean(a) for a in alignedData ] )
            if stdOfMean:
                stdDataList.append( [ scipy.std(a,ddof=1)/scipy.sqrt(len(a)) for a in alignedData ] )
            else:
                stdDataList.append( [ scipy.std(a,ddof=1) for a in alignedData ] )
            if returnAll:
                allDataList.append(alignedData)
        
        print ""
        meanDataListList.append(meanDataList)
        stdDataListList.append(stdDataList)
        timesList.append(allRelativeTimes)
        if returnAll: allDataListList.append(allDataList)

    if returnAll:
        return timesList,meanDataListList,stdDataListList,allDataListList
    else:
        return timesList,meanDataListList,stdDataListList

# 2/13/2017
def plotOutOfSampleVsTime(fileNumString,axListList=None,tmin=-1,tmax=3,
                          saveDir='./',savePrefix=None,laserBins=DEFAULTBINS,
                          legend=False,colors=BINCOLORS,**kwargs):
    timesList,meanDataListList,stdDataListList = \
        binnedOutOfSampleVsTime(fileNumString,laserBins=laserBins,**kwargs)
    
    if axListList is None: 
        pylab.figure(figsize=(4*len(timesList),3))
        axListList = [ None for i in range(len(timesList)) ]
    bins = laserBins
    
    if savePrefix is not None:
        saveFiles = []
    
    # loop over N
    for fignum,axList in enumerate(axListList):
        i = 0
        # loop over conditions
        for times,meanDataList,stdDataList in zip(timesList,meanDataListList,stdDataListList):
            i += 1
            if axList is None:
                ax = pylab.subplot(1,len(timesList),i)
            else:
                ax = axList[i-1]
            # loop over bins
            bini = 0
            for meanData,stdData in zip(meanDataList,stdDataList):
                bini += 1
                #pylab.errorbar(times,meanData,yerr=stdData)
                timeIndicesWithData = filter(lambda i: not scipy.isnan(meanData[i]),
                                             range(len(times)))
#                ax.plot(scipy.array(times)[timeIndicesWithData],
#                        scipy.array(meanData)[timeIndicesWithData],'-',
#                        label=str(bins[bini-1])+" to "+str(bins[bini]),
#                        color=colors[bini-1])
                defaultFigure.prettyErrorbar(ax,
                        scipy.array(times)[timeIndicesWithData],
                        scipy.array(meanData)[timeIndicesWithData],
                        scipy.array(stdData)[timeIndicesWithData],
                        label=str(int(round(bins[bini-1])))+"-"\
                             +str(int(round(bins[bini]))),
                        color=colors[bini-1])
                
                pylab.axis(xmin=tmin,xmax=tmax)
                pylab.xlabel('Time (s)')
                pylab.ylabel('Speed (px/s)')
                #pylab.axis(ymin=-20,ymax=20)
                if legend: leg = pylab.legend(loc=3,labelspacing=0.15)
                else: leg = None
                defaultFigure.makePretty(leg)
            #pylab.title(NAMES[i-1])
        pylab.subplots_adjust(wspace=0.3,left=0.175,right=0.95,bottom=0.175)

        if savePrefix is not None:
            saveFile = savePrefix+str(fignum)+'.pdf'
            pylab.figure(ax.figure.number)
            pylab.savefig(saveDir+saveFile)
            saveFiles.append(saveFile)

    if savePrefix is not None:
        return saveFiles

# 2/13/2017
def findLargestFitN(fpdData,fpdCombined,numFP,stopFittingN=None,ignoreFitAllDone=True):
    for N in scipy.sort(fpdData.keys())[::-1]:
            if fpdData[N]['fitAllDone'] or ignoreFitAllDone:
                if (N in fpdCombined) and hasattr(fpdCombined[N],'outOfSampleCostDict'):
                    if bestOutOfSampleNumStableFixedPoints(fpdCombined[N],
                                                           stopFittingN)[0]==numFP:
                        return N

# 2/13/2017
# 2.10.2017
def plotBinnedModelVsTime(fileNumString,laserBins=DEFAULTBINS,delay=10,
                          tmin=-1,tmax=3,ymin=-60,ymax=+30,Nskip=1,lw=1,ls='--',
                          var='speed',roundNum=4,forceAllTrials=False,
                          NtoTest=None,stopFittingN=3,fpd=None):
    fpdData = loadFitProbData(fileNumString)
    saveFilename = fpdData.values()[0]['saveFilename']
    fpdCombined = load(saveFilename[:-4]+'_combined.dat')
    numConditions = len(fpdData.values()[0]['fitProbDataList'])
    
    if NtoTest is None:
        # pick N to plot
        if numConditions == 1:
            # if there's only one condition, find largest N that has single fixed point solution
            N1fp = findLargestFitN(fpdData,fpdCombined,1,stopFittingN=stopFittingN)
            N2fp = findLargestFitN(fpdData,fpdCombined,2,stopFittingN=stopFittingN)
            NtoTest = [N1fp,N2fp]
        else:
            # otherwise plot largest 2 N that have been fit
            raise Exception, "10.12.2017 Need to fix to use stopFittingN"
            NtoTest = []
            sortedN = scipy.sort(fpdData.keys())[::-1]
            i = 0
            while len(NtoTest) < 2:
                if fpdData[sortedN[i]]['fitAllDone']:
                    if hasattr(fpdCombined[sortedN[i]],'outOfSampleCostDict'):
                        NtoTest.append(sortedN[i])
                i += 1
    print "NtoTest =",NtoTest
    
    axListList = []
    for N in NtoTest:
      if N is not None:
        #pylab.figure(figsize=(4*numConditions,3))
        
        axList = []
            
        timesList,meanDataListList,stdDataListList = binnedModelVsTime(fileNumString,
                N,laserBins=laserBins,Nskip=Nskip,roundNum=roundNum,
                forceAllTrials=forceAllTrials,var=var,tmin=tmin,tmax=tmax,delay=delay,
                stopFittingN=stopFittingN,fpd=fpd)
        
        # loop over conditions
        for conditioni in range(numConditions):
          times = timesList[conditioni]
          # loop over bins
          if len(BINCOLORS) + 1 < len(laserBins): raise Exception
          for bini,color in enumerate(BINCOLORS[:len(laserBins)-1]):
            meanModelData = meanDataListList[conditioni][bini]
            stdModelData = stdDataListList[conditioni][bini]
          
            ax = pylab.subplot(1,numConditions,conditioni+1)
            pylab.subplots_adjust(wspace=0.3,left=0.175,right=0.95,bottom=0.175)
            #pylab.plot(times,meanModelData,'--',lw=lw,color=color)
            defaultFigure.prettyErrorbar(ax,times,meanModelData,stdModelData,
                                         ls=ls,lw=lw,color=color)
            pylab.title(NAMES[conditioni]+', N='+str(N))
            pylab.axis(xmin=tmin,xmax=tmax,ymin=ymin,ymax=ymax)
            pylab.xlabel('Time (sec)')
            pylab.ylabel('Speed (pixels/sec)')
            defaultFigure.makePretty()
          axList.append(ax)
        axListList.append(axList)
        
    return axListList


# 6.27.2017
def plotBinnedModelVsTimeEnsemble(ensembleFilename,bincolors=BINCOLORS,**kwargs):
    """
    load and plot model ensemble data from generate_ensemble.ipynb
    
    confidence intervals represent the uncertainty in the model of the mean over inputs
    
    currently assumes there's only one condition
    """
    # data has shape (# ensemble members)x(#conditions)x(#bins)x(#inputs in bin)x(#times)
    dataDict = load(ensembleFilename)
    data = dataDict['allDataListListList']
    bins = dataDict['laserBins']
    times = dataDict['timesList'][0]
    
    # painfully make a list of numpy arrays 
    # (necessary because laser bins have different number of inputs per bin)
    # (remove condition axis since there's only one)
    # dataList has shape (#bins)x(#ensemble members)x(#inputs in bin)x(#times)
    conditioni = 0
    dataList = []
    for bini in range(len(bins)-1):
        databin = scipy.array(data)[:,conditioni,bini]
        databin = [ list(d) for d in databin ]
        databin = scipy.array(databin)
        dataList.append(databin)
        
    # now take mean over inputs in each bin
    # dataListMean has shape (#bins)x(#times)x(#ensemble members)
    dataListMean = scipy.array([ scipy.transpose(scipy.mean(d,axis=1)) for d in dataList ])
    
    # now take means and standard deviations of the means
    meanList = scipy.mean(dataListMean,axis=-1)
    stdList = scipy.std(dataListMean,axis=-1,ddof=1)
    lowerConfIntList = scipy.percentile(dataListMean,10,axis=-1)
    upperConfIntList = scipy.percentile(dataListMean,90,axis=-1)

    # make plot
    for mean,std,lower,upper,color in \
        zip(meanList,stdList,lowerConfIntList,upperConfIntList,bincolors):
        ax = pylab.gca()
        #defaultFigure.prettyErrorbar(ax,times,mean,std)
        defaultFigure.prettyConfInt(ax,times,mean,lower,upper,
                                    color=color,**kwargs)



def uniqueRelativeTimes(fileNumString,roundNum=4,forceAllTrials=False):
    """
    Returns unique relative times for each condition.
    Times are relative to the laser start, and do not include any delay.
    
    Returns array of shape (# conditions)x(# times)
    """

    testIndices,dataList,indepParamsListList = getOutOfSampleData(fileNumString,
                                                                  forceAllTrials)
    relativeTimesList = []
    
    # loop over conditions
    for data,indepParamsList in zip(dataList,indepParamsListList):
        # (independent parameter with index 0 is laser input, 1 is time delay)
        times = scipy.sort(data[0]['speed'].keys())-indepParamsList[0][1]
        
        allRelativeTimes = []
        for d,ip in zip(data,indepParamsList):
            allRelativeTimes.extend([ t-ip[1] for t in scipy.sort(d['speed'].keys()) ])
        
        relativeTimes = scipy.sort(scipy.unique(scipy.round_(allRelativeTimes,roundNum)))
        relativeTimesList.append(relativeTimes)

    return relativeTimesList


def binnedModelVsTime(fileNumString,N,laserBins=DEFAULTBINS,
                            Nskip=1,roundNum=4,forceAllTrials=False,
                            var='speed',tmin=-1.,tmax=3.,numpoints=100,
                            delay=10.,returnAll=False,parametersList=None,
                            times=None,stopFittingN=3,fpd=None):
    """
    Returns timesList,meanDataListList,stdDataListList.
    
    timesList has shape (#conditions)x(#times), and is measured relative to laser onset
    mean and stdDataListList have shape (#conditions)x(#bins)x(#times)
    
    stdData is the standard deviation of the mean.
    
    returnAll (False)           : Optionally return all individual trial data as
                                  a fourth return value, with shape 
                                  (#conditions)x(#bins)x(#inputs in that bin)x(#times)
    parametersList (None)       : If given, model parameters will be set as specified.
                                  Needs one set of parameters for each condition.
                                  (otherwise default to the loaded parameters)
    times (None)                : If given a list of relative times, use this instead
                                  of tmin,tmax,numpoints
    fpd (None)                  : Optionally explicitly pass fpd that contains
                                  the model.  In this case fileNumString will be used
                                  only to load indepParamsListList.
    """
    if fpd is None:
        loadIndividualFPs = True # old way of doing things below
        fpdData = loadFitProbData(fileNumString)
        numConditions = len(fpdData.values()[0]['fitProbDataList'])
        saveFilename = fpdData.values()[0]['saveFilename']
        fpdCombined = load(saveFilename[:-4]+'_combined.dat')
        fpd = fpdCombined[N]
    else:
        loadIndividualFPs = False # new way of doing things below
        numConditions = len(fpd.fittingProblemList)
    
    # only need indepParamsListList here...
    testIndices,dataList,indepParamsListList = getOutOfSampleData(fileNumString,
                                                                  forceAllTrials)
    
    timesList,meanDataListList,stdDataListList = [],[],[]
    if returnAll: allDataListList = []
    
    #relativeTimesList = uniqueRelativeTimes(fileNumString,roundNum,forceAllTrials)
    
    modelName = bestOutOfSampleModelName(fpd,stopFittingN)
    if modelName is None:
        raise Exception, "No best model: stopFittingN condition is not met"
    
    if times is None:
        times = scipy.linspace(delay+tmin,delay+tmax,numpoints)
        timesShifted = times - delay
    else:
        timesShifted = copy.copy(times)
        times = times + delay
    
    bins = scipy.array(laserBins)
    # loop over conditions
    for conditioni in range(numConditions):
        if loadIndividualFPs:
            print saveFilename,fileNumString,conditioni,N
            fp = loadFitProb(saveFilename,fileNumString,conditioni,N)
        else:
            fp = fpd.fittingProblemList[conditioni]
        print fp.fittingModelDict.keys()
        model = fp.fittingModelDict[modelName]
        if parametersList is not None: # optionally set parameters
            if len(model.getParameters()) != len(parametersList[conditioni]):
                raise Exception
            model.initializeParameters(parametersList[conditioni])
        
        laserList = scipy.array(indepParamsListList[conditioni])[:,0]
        delayList = scipy.array(indepParamsListList[conditioni])[:,1]
        meanDataList,stdDataList,allDataList = [],[],[]
        # loop over laser bins
        for binmin,binmax in zip(bins[:-1],bins[1:]):
            numPerBin = 0
            modelData = []
            # loop over experimental trials
            for laser in laserList[::Nskip]:
              #absTimes = relativeTimes
              if binmin <= laser < binmax:
                numPerBin += 1
                oneInputModelData = model.evaluateVec(times,var,[laser,delay])
                modelData.append( oneInputModelData )
            meanData = scipy.mean(modelData,axis=0)
            stdData = scipy.std(modelData,axis=0,ddof=1)/scipy.sqrt(len(modelData))
            meanDataList.append(meanData)
            stdDataList.append(stdData)
            if returnAll: allDataList.append(modelData)
            print "binnedModelVsTime: numPerBin =",numPerBin
        
        print ""
        meanDataListList.append(meanDataList)
        stdDataListList.append(stdDataList)
        timesList.append(timesShifted)
        if returnAll: allDataListList.append(allDataList)
        
    if returnAll:
        return timesList,meanDataListList,stdDataListList,allDataListList
    else:
        return timesList,meanDataListList,stdDataListList

# 2/13/2017 adapting from SloppyCell.IO.eqns_TeX_file
def pdfReport(fileNumString,Nskip=1,forceAllTrials=True,tmin=-1.,tmax=4.,
              laserBins=DEFAULTBINS,stopFittingN=3,runCombine=True):
    
    if runCombine: # needs to be run whenever additional data is ready
        combineFitProbs(fileNumString)
    
    dir = fileNumString+"_reports/"
    
    # () make and save figure pdfs
    figFiles = []
    
    # get number of in-sample trials
    config = load(fileNumString+"_configDict.dat")
    numInSampleTrials = config['numInSampleTrials']
    saveFilename = loadFitProbData(fileNumString).values()[0]['saveFilename']
    allDataFilename = saveFilename[:-4]+'_allData.dat'
    allData = load(allDataFilename)
    inSampleTrialsList = []
    for fp in allData.fittingProblemList:
        if numInSampleTrials is None:
            inSampleTrialsList.append( len(fp.fittingData) )
        else:
            inSampleTrialsList.append( numInSampleTrials )
    if len(pylab.unique(inSampleTrialsList)) == 1:
        numInSampleTrials = inSampleTrialsList[0]
    else:
        numInSampleTrials = None
        print "6.13.2017: WARNING: different numInSampleTrials in different conditions"
    
    # goodness of fit
    fitProbData = loadFitProbData(fileNumString)
    saveFilename = fitProbData.values()[0]['saveFilename']
    combinedFilename = saveFilename[:-4]+'_combined.dat'
    fpdCombined = load(combinedFilename)
    fig2name = fileNumString+'_goodnessOfFit.pdf'
    plotGoodnessOfFit(fpdCombined,numInSampleTrials=numInSampleTrials,
                      stopFittingN=stopFittingN)
    pylab.savefig(dir+fig2name)
    figFiles.append(fig2name)
    
    # binned model and experimental speed versus time
    fig1prefix = fileNumString+'_binnedVsTime'
    axListList = plotBinnedModelVsTime(fileNumString,Nskip=Nskip,
                                       tmin=tmin,tmax=tmax,
                                       laserBins=laserBins,
                                       forceAllTrials=forceAllTrials,
                                       stopFittingN=stopFittingN)
    figFiles.extend( plotOutOfSampleVsTime(fileNumString,
                                     axListList=axListList,
                                     savePrefix=fig1prefix,
                                     saveDir=dir,
                                     tmin=tmin,tmax=tmax,
                                     forceAllTrials=forceAllTrials,
                                     laserBins=laserBins) )
    
    # () make tex file
    lines = []
    
    lines.append(r'\documentclass{article}')
    lines.append(r'\usepackage{amsmath}')
    lines.append(r'\usepackage{fullpage}')
    lines.append(r'\usepackage{graphicx}')
    lines.append(r'\title{'+fileNumString+'}')
    lines.append(r'\begin{document}')
    lines.append(r'\maketitle')
    lines.append(r'\pagenumbering{gobble}')
    
    # () include info about in-sample and out-of-sample indices
    # number of in-sample trials
    lines.append(r'Number of in-sample trials:')
    lines.append(str(inSampleTrialsList))
    lines.append(r'')
    lines.append(r'')
    # out-of-sample ranges
    finishedSubsets = countFitProbData(fileNumString,printReport=False)
    outOfSampleRangesList = []
    for numTimepoints in finishedSubsets:
        if numTimepoints in fpdCombined:
          fp = fpdCombined[numTimepoints]
          if hasattr(fp,'outOfSampleCostIndices'):
            outOfSampleRanges = \
                [ [indices[0],indices[-1]] for indices in fp.outOfSampleCostIndices ]
            outOfSampleRangesList.append(str(outOfSampleRanges))
    uniqueOutOfSampleRanges = scipy.unique(outOfSampleRangesList)
    if len(uniqueOutOfSampleRanges) > 1:
        lines.append(r'WARNING: Multiple out-of-sample ranges used.')
        lines.append(r'')
        lines.append(r'')
    for outOfSampleRange in uniqueOutOfSampleRanges:
        lines.append(r'Out-of-sample ranges:')
        lines.append(outOfSampleRange)
        lines.append(r'')
        lines.append(r'')

    # note stopFittingN
    lines.append(r'stopFittingN = '+str(stopFittingN))
    lines.append(r'')
    lines.append(r'')
    
    # make single figure that includes multiple plots
    lines.append(r'\begin{figure}[h]')
    lines.append(r'\centering')
    for figFile in figFiles:
        lines.append(r'\includegraphics[scale=0.55]{'+figFile+'}')
    lines.append(r'\end{figure}')
    
    lines.append(r'\end{document}')

    filename = '%s.tex' % fileNumString
    f = file(dir+filename, 'w')
    f.write(os.linesep.join(lines))
    f.close()

    # () render pdf
    call(['pdflatex',filename],cwd=dir)

# 10.3.2017
def fitProbDataOutOfSampleFilename(fileNumString):
    return fileNumString+'_fitProbDataOutOfSample.dat'

# 10.3.2017
def loadFitProbDataOutOfSample(fileNumString):
    """
    Loads fitProbData file that contains out-of-sample cost data.
    """
    return load(fitProbDataOutOfSampleFilename(fileNumString))

# 10.3.2017
def saveFitProbDataOutOfSample(fitProbDataOutOfSample,fileNumString):
    save(fitProbDataOutOfSample,fitProbDataOutOfSampleFilename(fileNumString))

# 10.3.2017
def setupFitProbDataOutOfSample(fileNumString):
    """
    If it is not already present, set up a fitProbData file that 
    contains out-of-sample cost data.
    """
    if not os.path.isfile(fitProbDataOutOfSampleFilename(fileNumString)):
        fpd = loadFitProbData(fileNumString)
        fpdOutOfSample = copy.deepcopy(fpd)
        for N in fpdOutOfSample.keys():
            fpdOutOfSample[N]['outOfSampleCostDict'] = {}
            fpdOutOfSample[N].pop('fitAllDone')
            fpdOutOfSample[N].pop('logLikelihoodDict')
            for f in fpdOutOfSample[N]['fitProbDataList']:
                f.pop('fittingStateDict')
                f.pop('logLikelihoodDict')
        saveFitProbDataOutOfSample(fpdOutOfSample,fileNumString)

# 10.3.2017
def updateFitProbDictOutOfSample(fp,N,fileNumString):
    fpdOutOfSample = loadFitProbDataOutOfSample(fileNumString)
    fpdOutOfSample[N]['outOfSampleCostDict'] = fp.outOfSampleCostDict
    saveFitProbDataOutOfSample(fpdOutOfSample,fileNumString)


# 10.3.2017
def subsetsThatNeedOutOfSampleCost(fileNumString,includeAll=False):
    """
    Find data subsets (N) that have models that have been fit 
    but whose out-of-sample cost has not been calculated.
    
    includeAll (False)      : Optionally include all subsets with
                              fit models, even if they already have
                              out of sample costs calculated
    """
    fpd = loadFitProbData(fileNumString)
    fpdOutOfSample = loadFitProbDataOutOfSample(fileNumString)

    Nlist = []
    for N in scipy.sort(fpd.keys()):
        fitModels = fpd[N]['logLikelihoodDict'].keys()
        if includeAll:
            if len(fitModels) > 0:
                Nlist.append(N)
        else:
            costCalculatedModels = fpdOutOfSample[N]['outOfSampleCostDict'].keys()
            if len(scipy.intersect1d(fitModels,costCalculatedModels)) < len(fitModels):
                Nlist.append(N)

    return Nlist

# 10.4.2017
def filenameNonly(fileNumString,numTimepoints):
    saveFilename = loadFitProbData(fileNumString).values()[0]['saveFilename']
    return directoryPrefixNonly(fileNumString,numTimepoints)+'/'+saveFilename

# 10.4.2017
def loadCombinedFittingProblem(fileNumString,numTimepoints):
    Nfilename = filenameNonly(fileNumString,numTimepoints)
    try:
        fp = load(Nfilename)
        return fp
    except IOError:
        # file may not be there if fitProbData file is inconsistent
        # with the current state of the rest of the filesystem
        print "gatherResults: Unable to load",Nfilename
        return None

# 10.4.2017
def saveCombinedFittingProblem(fp,fileNumString,numTimepoints):
    Nfilename = filenameNonly(fileNumString,numTimepoints)
    save(fp,Nfilename)

if __name__ == '__main__':

    # index of the fitting problem database file is set by command-line argument
    if len(sys.argv) < 2 or len(sys.argv) > 2:
        print "Usage: python gatherResults.py [fileNumString]"
        exit()
    else:
        fileNumString = sys.argv[1]

    #saveFilename = loadFitProbData(fileNumString).values()[0]['saveFilename']
    #combinedFilename = saveFilename[:-4]+'_combined.dat'

    testIndices,dataList,indepParamsListList = getOutOfSampleData(fileNumString)

    if RECALCULATECOST:
        combineFitProbs(fileNumString,reset=True)

    # make out-of-sample fitting problem database file if it doesn't exist
    setupFitProbDataOutOfSample(fileNumString)

    #finishedSubsetsBefore = []
    while True:
        
        subsetsToRun = subsetsThatNeedOutOfSampleCost(fileNumString,
                                                      includeAll=REMAKEREPORTS)
        REMAKEREPORTS = False
        
        if len(subsetsToRun) > 0:
            combineFitProbs(fileNumString,saveCombined=False)
            #fpd = load(combinedFilename)
        
            for numTimepoints in subsetsToRun[::-1]:
                #fp = fpd[numTimepoints]
                fp = loadCombinedFittingProblem(fileNumString,numTimepoints)
                    
                if fp is not None:
                    
                    print "gatherResults: Calculating out-of-sample costs for N =",numTimepoints,"..."
                    
                    # () calculate out-of-sample goodness of fit (chisq)
                    fp.outOfSampleCostIndices = testIndices
                    updateGoodnessOfFit(fp,dataList,indepParamsListList)
                    # save chisq data in fitProbDictOutOfSample
                    updateFitProbDictOutOfSample(fp,numTimepoints,fileNumString)
                    # save updated fp within directory structure
                    saveCombinedFittingProblem(fp,fileNumString,numTimepoints)
                
                    # () find best model based on out-of-sample predictions
                    bestModelName = bestOutOfSampleModelName(fp,STOPFITTINGN)
                    
                    if bestModelName is not None:
                        print "gatherResults: Making reports for N =",numTimepoints,",",bestModelName
                    
                        # () make plots, etc.
                        wormReport(fileNumString,numTimepoints,bestModelName)

                        # () store best model fixed points in fp
                        # (read from Mathematica file)
                        storeFixedPoints(fileNumString,numTimepoints,bestModelName)


    
        time.sleep(60) # run once per minute
        #finishedSubsetsBefore = copy.copy(finishedSubsets)




