# pulseInput.py
#
# Bryan Daniels
# 4.28.2016
#
# Modify exiting fittingModels to include input that consists of a
# transient pulse instead of a constant over time.
#

from SirIsaac.FittingProblem import *

def pulseInput(fittingProblem,inputName,pulseDuration,dynamicsType='pulse',
    delay=0.,delayOptimizable=False,delayAsInput=False,tau=1.,tauOptimizable=False,
    delayList=None):
    """
    Modify existing fittingModels to include input that consists of a
    transient pulse instead of a constant over time.
    
    fittingModelList must be a list of SloppyCellFittingModels.
    
    Only works for input values that appear solely on the righthand side
    of the fittingModel's rate rules or assignment rules.  (I assume this
    should cover most cases...)
    
    dynamicsType ("pulse")          : "pulse" --> Step pulse
                                      "riseAndDecay" --> Linear rise during pulse, then
                                            exponential decay with timescale tau
    delayOptimizable (False)        : If True, delay is an optimizable parameter.
    delayAsInput (False)            : If True, delay is set up as an independent 
                                      parameter (input) to the model.
                                      Use delayList kwarg to pass a list of delays that will
                                      be added to the fittingProblem's indepParamsList.
                                      (delayList has the effect of overriding the 
                                      "delay" kwarg)
    """
    if delayOptimizable and delayInput: raise Exception
    
    fittingModelList = fittingProblem.fittingModelList
    
    for fittingModel in fittingModelList:
        if inputName not in fittingModel.indepParamNames: raise Exception
    
        net = fittingModel.net
        if len(net.compartments) > 1: raise Exception
        comp = net.compartments[0].id

        # if the input name is contained in the name of any other existing
        # variable (or if it's not an existing variable), we're going to have trouble
        numMatching = 0
        for existingName in net.get_var_vals().keys():
            if existingName.find(inputName) > -1: numMatching += 1
        if numMatching != 1: raise Exception

        # create new species that will be the time varying input,
        # and insert it into the network in place of the original input
        timeInputName = inputName+"_timeVarying"
        net.add_species(timeInputName,comp,typical_value=1.)

        # change existing references to the original input in the rate
        # rules and assignment rules
        for rateLHS in net.rateRules.keys():
            rateRHS = net.rateRules.get(rateLHS)
            newRateRHS = rateRHS.replace(inputName,timeInputName)
            net.addRateRule(rateLHS,newRateRHS)
        for assignLHS in net.assignmentRules.keys():
            assignRHS = net.assignmentRules.get(assignLHS)
            newAssignRHS = assignRHS.replace(inputName,timeInputName)
            net.addAssignmentRule(assignLHS,newAssignRHS)

        # update model's species names
        fittingModel.speciesNames.append(timeInputName)

        # specify the transient dynamics
        delayInputName = inputName+"_delay"
        net.addParameter(delayInputName,value=delay,
                         isOptimizable=delayOptimizable,
                         typicalValue=delay)
        if dynamicsType is "pulse":
            net.add_event(inputName+'On',
                          'gt(time, '+delayInputName+')',
                          {timeInputName: inputName})
            net.add_event(inputName+'Off',
                          'gt(time, '+delayInputName+' + '+str(pulseDuration)+')',
                          {timeInputName: 0.})
        elif dynamicsType is "riseAndDecay":
            # linear rise
            net.addParameter('F1',value=0.,
                             isOptimizable=False,
                             typicalValue=1.,
                             isConstant=False)
            net.add_assignment_rule('F1',
                '(time-'+delayInputName+')/'+str(pulseDuration))
            # exponential decay
            net.addParameter('F2',value=0.,
                             isOptimizable=False,
                             typicalValue=1.,
                             isConstant=False)
            net.addParameter('tau_pulse',value=tau,
                             isOptimizable=tauOptimizable,
                             typicalValue=tau)
            net.add_assignment_rule('F2',
                'exp( -(time-'+delayInputName+'-'+str(pulseDuration)+')/tau_pulse )')
            # switch between the two
            net.addParameter('Theta1',value=0.,
                             isOptimizable=False,
                             typicalValue=1.,
                             isConstant=False)
            net.addParameter('Theta2',value=0.,
                              isOptimizable=False,
                             typicalValue=1.,
                             isConstant=False)
            net.add_assignment_rule(timeInputName,inputName+'*(Theta1*F1 + Theta2*F2)')
            net.add_event(inputName+'On',
                          'gt(time, '+delayInputName+')',
                          {'Theta1': 1,
                           'Theta2': 0})
            net.add_event(inputName+'Off',
                          'gt(time, '+delayInputName+' + '+str(pulseDuration)+')',
                          {'Theta1': 0,
                           'Theta2': 1})
        else:
            raise Exception, "Unrecognized dynamicsType"


        if delayAsInput: # add delay as an input parameter
            
            fittingModel.indepParamNames.append(delayInputName)
            fittingModel.numInputs += 1

        # TO DO: update model's netList
        # (check that you get the right diagram using networkFigureBestModel)


    if delayAsInput:
        # update fittingProblem's indepParamNames (if it isn't already)
        if delayInputName not in fittingProblem.indepParamNames:
            fittingProblem.indepParamNames.append(delayInputName)
        
        # append delay for each trial to indepParamsList
        if (delayList is None) \
            or (len(delayList) != len(fittingProblem.indepParamsList)):
            raise Exception
        for d,indepParams in zip(delayList,fittingProblem.indepParamsList):
            indepParams.append(d)
            


        